﻿// <copyright file="DcClient.cs" company="Mek">
// DC Search by Mek (www.mekweb.eu), all rights reserved.
// </copyright>
// <author>Mek</author>

namespace DcLibrary
{
    using System;
    using System.Collections.Generic;
    using System.Globalization;
    using System.IO;
    using System.Linq;
    using System.Net;
    using System.Net.Sockets;
    using System.Text;
    using System.Threading;

    /// <summary>
    /// Represents a DirectConnect client
    /// </summary>
    public class DcClient
    {
        /// <summary>
        /// Nick template to search the hubs under. Random characters are appended to it when necessary.
        /// </summary>
        private const string NickTemplate = "zegepard0";

        /// <summary>
        /// Internal list of DC clients to reuse if possible
        /// </summary>
        private static IList<DcClient> clients = new List<DcClient>();

        /// <summary>
        /// Stores IP address to listen on for UDP datagrams with search results
        /// </summary>
        private string listenIPAddress;

        /// <summary>
        /// Stores UDP port number to listen on for datagrams with search results
        /// </summary>
        private int listenUdpPort;

        /// <summary>
        /// Stores search parameters
        /// </summary>
        private SearchParameters searchParameters;

        /// <summary>
        /// Stores the callback to invoke when results from search are obtained
        /// </summary>
        private Action<IEnumerable<SearchResult>> gotSearchResultsAction;

        /// <summary>
        /// Stores the callback to invoke when client is connecting to a hub
        /// </summary>
        private Action<Hub, int, int> connectingToHubAction;

        /// <summary>
        /// Stores the callback to invoke when connection to a hub failed
        /// </summary>
        private Action<string> connectionToHubFailedAction;

        /// <summary>
        /// Stores the callback to invoke when search operation is finished
        /// </summary>
        private Action searchFinishedAction;

        /// <summary>
        /// Stores list of hubs to perform batch search on
        /// </summary>
        private IEnumerable<Hub> hubs;

        /// <summary>
        /// Stores the thread that performs batch searching
        /// </summary>
        private Thread batchSearchThread;

        /// <summary>
        /// Stores currently used nick generated from NickTemplate
        /// </summary>
        private string nick;

        /// <summary>
        /// Initializes a new instance of the <see cref="DcClient" /> class
        /// </summary>
        /// <param name="listenIPAddress"></param>
        /// <param name="listenUdpPort"></param>
        public DcClient(string listenIPAddress, ushort listenUdpPort)
        {
            if (!StringUtility.IsIPAddress(listenIPAddress))
            {
                throw new ArgumentException("Invalid listen IP address.", "listenIPAddress");
            }

            if (listenUdpPort > 65535)
            {
                throw new ArgumentException("Listen UDP port must be between 0 and 65535.", "listenUdpPort");
            }

            if (clients.Any(c => c.ListenIPAddress == listenIPAddress && c.ListenUdpPort == listenUdpPort))
            {
                throw new ArgumentException("Client with provided listenIPAddress and listenUdpPort already exists");
            }

            this.listenIPAddress = listenIPAddress;
            this.listenUdpPort = listenUdpPort;
        }

        /// <summary>
        /// Gets IP address to listen on for UDP datagrams with search results
        /// </summary>
        public string ListenIPAddress
        {
            get { return this.listenIPAddress; }
        }

        /// <summary>
        /// Gets UDP port to listen on for datagrams with search results
        /// </summary>
        public int ListenUdpPort
        {
            get { return this.listenUdpPort; }
        }

        /// <summary>
        /// Gets MyInfo string to identify client against the hub
        /// </summary>
        private string MyInfo
        {
            get
            {
                return "$MyINFO $ALL " + this.nick + " Nobody.<++ V:0.871,M:A,H:2/0/0,S:5>$ $5\x08$never@write.me$217084659904$|";
            }
        }

        /// <summary>
        /// Starts batch search
        /// </summary>
        /// <param name="hublistXmlContent"></param>
        /// <param name="parameters"></param>
        /// <param name="gotSearchResultsAction"></param>
        /// <param name="connectingToHubAction"></param>
        /// <param name="connectionToHubFailedAction"></param>
        /// <param name="searchFinishedAction"></param>
        public void StartBatchSearch(
            string hublistXmlContent,
            SearchParameters parameters,
            Action<IEnumerable<SearchResult>> gotSearchResultsAction,
            Action<Hub, int, int> connectingToHubAction,
            Action<string> connectionToHubFailedAction,
            Action searchFinishedAction)
        {
            this.searchParameters = parameters;
            this.gotSearchResultsAction = gotSearchResultsAction;
            this.connectingToHubAction = connectingToHubAction;
            this.connectionToHubFailedAction = connectionToHubFailedAction;
            this.searchFinishedAction = searchFinishedAction;

            this.nick = NickTemplate + DateTime.Now.Second;
            this.hubs = HublistParser.Parse(hublistXmlContent);
            this.batchSearchThread = new Thread(new ThreadStart(this.BatchSearchWorker));
            this.batchSearchThread.IsBackground = true;
            this.batchSearchThread.Start();
        }

        /// <summary>
        /// Stops batch search
        /// </summary>
        public void StopBatchSearch()
        {
            try
            {
                this.batchSearchThread.Abort();
            }
            catch (ThreadAbortException)
            {
            }
        }

        /// <summary>
        /// Represents work for the batch search thread
        /// </summary>
        private void BatchSearchWorker()
        {
            var udpEndpointLocal = new IPEndPoint(IPAddress.Any, this.listenUdpPort);
            var udpEndpointRemote = (EndPoint)new IPEndPoint(IPAddress.Any, 0);

            // for debugging
            // this.hubs = new List<Hub>() { new Hub() { Address = "tranceworld.cz", Port = 411, UserCount = 111 } };
            int currentCount = 0;

            foreach (Hub hub in this.hubs)
            {
                using (var tcpClient = new TcpClient())
                {
                    currentCount++;

                    if (this.connectingToHubAction != null)
                    {
                        this.connectingToHubAction(hub, currentCount, this.hubs.Count());
                    }

                    try
                    {
                        tcpClient.Connect(hub.Address, hub.Port);
                    }
                    catch (SocketException)
                    {
                        if (this.connectionToHubFailedAction != null)
                        {
                            this.connectionToHubFailedAction("Could not connect");
                            continue;
                        }
                    }

                    using (var tcpStream = tcpClient.GetStream())
                    {
                        int commandWaitCount = 0;

                        while (true)
                        {
                            var commands = this.GetCommands(tcpStream).Where(c => !string.IsNullOrWhiteSpace(c));
                            bool done = false;

                            foreach (string command in commands)
                            {
                                if (command.StartsWith("$Lock"))
                                {
                                    string key = StringUtility.Lock2Key(command);
                                    this.WriteToStream("$Supports NoGetINFO|$Key " + key + "|$ValidateNick " + this.nick + "|", tcpStream);
                                }
                                else if (command.StartsWith("$Hello") && !done)
                                {
                                    done = true;
                                    this.WriteToStream("$Version 1,0091|$GetNickList|" + this.MyInfo, tcpStream);
                                    Thread.Sleep(1000);

                                    // start listening for search results
                                    using (var udpSocket = new Socket(AddressFamily.InterNetwork, SocketType.Dgram, ProtocolType.Udp))
                                    {
                                        udpSocket.Bind(udpEndpointLocal);

                                        // wait 10s so the hub does not consider us spammers
                                        Thread.Sleep(10000);

                                        // send search request
                                        var searchString = string.Format(
                                            CultureInfo.InvariantCulture,
                                            "$Search {0}:{1} {2}?{3}?{4}?{5}?{6}|",
                                            this.listenIPAddress,
                                            this.listenUdpPort,
                                            this.searchParameters.SizeConstraintType == SizeConstraintType.None ? "F" : "T",
                                            this.searchParameters.SizeConstraintType == SizeConstraintType.AtLeast ? "F" : "T",
                                            this.searchParameters.SizeConstraintType == SizeConstraintType.None ? "0" : this.searchParameters.Size.ToString(CultureInfo.InvariantCulture),
                                            (int)this.searchParameters.FileType,
                                            StringUtility.SanitizeSearchString(this.searchParameters.Text));

                                        this.WriteToStream(searchString, tcpStream);

                                        // wait 5s for results
                                        Thread.Sleep(5000);

                                        // collect results
                                        StringBuilder sb = new StringBuilder();
                                        byte[] readBuffer = new byte[1024];
                                        int byteCount;

                                        while (udpSocket.Available != 0)
                                        {
                                            byteCount = udpSocket.ReceiveFrom(readBuffer, ref udpEndpointRemote);
                                            sb.Append(Encoding.Default.GetString(readBuffer, 0, byteCount));
                                        }

                                        // file: $SR matys666 MP3\Electronic\midilidi\03 Lodnici.mp314134556 1/3TTH:AJLTCKBKNR3VN6Y73S5RHC5JAPCSWRRNLFY3GRY (87.236.195.19:999)|
                                        // folder: $SR Mek Mobil\Zvuky\MIDI 2/2TTH:GMAAAAFDMOTQGVASJMADGAAAACMGHJYDTBM3QAI (87.236.195.19:999)|
                                        var resultsRaw = sb.ToString().Split('|')
                                            .Where(s => !string.IsNullOrWhiteSpace(s))
                                            .Select(s => s.Substring(4)); // remove $SR

                                        var results = from resultRaw in resultsRaw
                                                      let resultsSplit = resultRaw.Split('\x5')
                                                      where resultsSplit.Count() > 1
                                                      let isFolder = resultsSplit.Count() == 2
                                                      let nameWithPath = isFolder
                                                          ? resultsSplit[0].Substring(resultsSplit[0].IndexOf(' ') + 1, resultsSplit[0].LastIndexOf(' ') - resultsSplit[0].IndexOf(' ') - 1)
                                                          : resultsSplit[0].Substring(resultsSplit[0].IndexOf(' ') + 1)
                                                      select new SearchResult()
                                                      {
                                                          HubAddress = hub.Address + ":" + hub.Port,
                                                          IsFolder = isFolder,
                                                          Name = nameWithPath.Contains('\\') ? nameWithPath.Substring(nameWithPath.LastIndexOf('\\') + 1) : nameWithPath,
                                                          Nick = resultsSplit[0].Substring(0, resultsSplit[0].IndexOf(' ')),
                                                          Path = nameWithPath.Contains('\\') ? nameWithPath.Substring(0, nameWithPath.LastIndexOf('\\')) : string.Empty,
                                                          Size = isFolder
                                                              ? default(long)
                                                              : long.Parse(resultsSplit[1].Split(' ')[0]),
                                                          Slots = isFolder
                                                              ? resultsSplit[0].Substring(resultsSplit[0].LastIndexOf(' ') + 1)
                                                              : resultsSplit[1].Split(' ')[1],
                                                          TTH = isFolder
                                                              ? string.Empty
                                                              : resultsSplit[2].Substring(4, 39)
                                                      };

                                        if (this.searchParameters.FreeSlotsOnly)
                                        {
                                            results = results.Where(r => !r.Slots.StartsWith("0/"));
                                        }

                                        if (this.gotSearchResultsAction != null)
                                        {
                                            this.gotSearchResultsAction(results);
                                        }

                                        // go to next hub
                                        break;
                                    }
                                }
                                else if (command.StartsWith("$ForceMove"))
                                {
                                    if (this.connectionToHubFailedAction != null)
                                    {
                                        this.connectionToHubFailedAction("Hub is redirecting");
                                        done = true;
                                        break;
                                    }
                                }
                                else if (command.StartsWith("$ValidateDenide"))
                                {
                                    if (this.connectionToHubFailedAction != null)
                                    {
                                        this.connectionToHubFailedAction("Login refused");
                                        done = true;
                                        break;
                                    }
                                }
                                else if (command.StartsWith("$GetPass"))
                                {
                                    if (this.connectionToHubFailedAction != null)
                                    {
                                        this.connectionToHubFailedAction("Password required");
                                        done = true;
                                        break;
                                    }
                                }
                            }

                            if (!commands.Any() && !done)
                            {
                                // no commands, wait for a while and maybe something appears later
                                commandWaitCount++;

                                if (commandWaitCount > 10)
                                {
                                    break;
                                }

                                Thread.Sleep(1000);
                            }

                            if (done)
                            {
                                break;
                            }
                        }
                    }
                }
            }

            if (this.searchFinishedAction != null)
            {
                this.searchFinishedAction();
            }
        }

        /// <summary>
        /// Reads from provided stream and parses the string coming from it into DirectConnect commands
        /// </summary>
        /// <param name="stream"></param>
        /// <returns></returns>
        private IEnumerable<string> GetCommands(NetworkStream stream)
        {
            StringBuilder sb = new StringBuilder();
            byte[] readBuffer = new byte[1024];
            int byteCount;

            while (stream.DataAvailable)
            {
                byteCount = stream.Read(readBuffer, 0, readBuffer.Length);
                sb.Append(Encoding.Default.GetString(readBuffer, 0, byteCount));
            }

            return sb.ToString().TrimEnd('\0').Split('|');
        }

        /// <summary>
        /// Writes specified input to specified stream
        /// </summary>
        /// <param name="input"></param>
        /// <param name="stream"></param>
        private void WriteToStream(string input, Stream stream)
        {
            var bytes = Encoding.Default.GetBytes(input);

            try
            {
                stream.Write(bytes, 0, bytes.Length);
            }
            catch (IOException)
            {
            }
        }
    }
}
