﻿// <copyright file="HublistParser.cs" company="Mek">
// DC Search by Mek (www.mekweb.eu), all rights reserved.
// </copyright>
// <author>Mek</author>

namespace DcLibrary
{
    using System.Collections.Generic;
    using System.Linq;
    using System.Text;
    using System.Xml.Linq;

    /// <summary>
    /// Class for parsing a hublist
    /// </summary>
    internal static class HublistParser
    {
        /// <summary>
        /// Parses hub information from the provided stream that should contain a XML document
        /// </summary>
        /// <param name="hublistXmlContent"></param>
        /// <returns></returns>
        internal static IEnumerable<Hub> Parse(string hublistXmlContent)
        {
            XDocument xmlDocument = XDocument.Parse(XmlCharacterWhitelist(hublistXmlContent), LoadOptions.None);

            var hubElements = xmlDocument.Descendants(XName.Get("Hubs"))
                .Descendants(XName.Get("Hub"));

            var result = from hubElement in hubElements
                         let address = hubElement.Attribute(XName.Get("Address")).Value
                         let addressLastSlashIndex = address.LastIndexOf('/') != -1
                            ? address.LastIndexOf('/') + 1 // address with protocol
                            : 0 // address without protocol
                         let addressLastColonIndex = address.LastIndexOf(':') + 1
                         where !address.StartsWith("adc://") && !address.StartsWith("adcs://") && !address.StartsWith("nmdcs://")
                         where address != "dchub://support.dchublist.com" // bogus address
                         select new Hub()
                {
                    Address = addressLastColonIndex > addressLastSlashIndex
                        ? address.Substring(addressLastSlashIndex, addressLastColonIndex - addressLastSlashIndex - 1) // address with port
                        : address.Substring(addressLastSlashIndex), // address without port
                    Port = addressLastColonIndex > addressLastSlashIndex
                        ? ushort.Parse(address.Substring(addressLastColonIndex)) // address with port
                        : (ushort)411, // address without port
                    UserCount = ushort.Parse(hubElement.Attribute(XName.Get("Users")).Value)
                };

            return result;
        }

        /// <summary>
        /// Sanitizes XML string from invalid characters
        /// </summary>
        /// <param name="input"></param>
        /// <returns></returns>
        private static string XmlCharacterWhitelist(string input)
        {
            if (input == null) return null;

            StringBuilder sbOutput = new StringBuilder();
            char ch;

            for (int i = 0; i < input.Length; i++)
            {
                ch = input[i];
                if ((ch >= 0x0020 && ch <= 0xD7FF) ||
                    (ch >= 0xE000 && ch <= 0xFFFD) ||
                    ch == 0x0009 ||
                    ch == 0x000A ||
                    ch == 0x000D)
                {
                    sbOutput.Append(ch);
                }
            }

            return sbOutput.ToString();
        }
    }
}
