﻿// <copyright file="HublistDownloader.cs" company="Mek">
// DC Search by Mek (www.mekweb.eu), all rights reserved.
// </copyright>
// <author>Mek</author>

namespace DcSearch
{
    using System;
    using System.IO;
    using System.Net;
    using ICSharpCode.SharpZipLib.BZip2;

    /// <summary>
    /// Class that allows download of a hublist
    /// </summary>
    public static class HublistDownloader
    {
        /// <summary>
        /// Downloads .XML.BZ2 hublist from specified address to specified path, with optional callback
        /// </summary>
        /// <param name="hublistAddress"></param>
        /// <param name="hublistPath"></param>
        /// <param name="callback"></param>
        public static void StartHublistDownload(string hublistAddress, string hublistPath, Action callback)
        {
            if (callback != null)
            {
                WebClient webClient = new WebClient()
                {
                    Proxy = null
                };

                webClient.DownloadDataAsync(new Uri(hublistAddress));

                DownloadDataCompletedEventHandler completedHandler = null;

                completedHandler = (s, e) =>
                {
                    webClient.DownloadDataCompleted -= completedHandler;

                    if (!e.Cancelled && e.Error == null)
                    {
                        using (var outStream = new MemoryStream())
                        using (var inStream = new MemoryStream(e.Result))
                        {
                            BZip2.Decompress(inStream, outStream, false);

                            string hublistDir = Path.GetDirectoryName(hublistPath);
                            if (!Directory.Exists(hublistDir))
                            {
                                Directory.CreateDirectory(hublistDir);
                            }
                            
                            using (var reader = new BinaryReader(outStream))
                            using (var fileStream = File.Create(hublistPath))
                            {
                                reader.BaseStream.Seek(0, SeekOrigin.Begin);

                                while (outStream.Position != outStream.Length)
                                {
                                    byte readByte = reader.ReadByte();
                                    
                                    // if there is illegal XML character in name of hub, strip it out
                                    if (readByte > 31)
                                    {
                                        fileStream.WriteByte(readByte);
                                    }
                                }
                            }

                            // overcome NTFS creation time caching by setting it explicitly everytime
                            File.SetCreationTime(hublistPath, DateTime.Now);
                        }
                        
                        if (callback != null)
                        {
                            callback();
                        }
                    }
                };

                webClient.DownloadDataCompleted += completedHandler;
            }
        }
    }
}
