﻿// <copyright file="MainForm.cs" company="Mek">
// DC Search by Mek (www.mekweb.eu), all rights reserved.
// </copyright>
// <author>Mek</author>

namespace DcSearch
{
    using System;
    using System.Collections.ObjectModel;
    using System.Diagnostics.CodeAnalysis;
    using System.Drawing;
    using System.Globalization;
    using System.IO;
    using System.Linq;
    using System.Windows.Forms;
    using DcLibrary;
    using System.Text;

    /// <summary>
    /// Main application form
    /// </summary>
    public partial class MainForm : Form
    {
        /// <summary>
        /// Character that, when appended to log text, makes the text red
        /// </summary>
        private const char ErrorIndicator = '\x01';

        /// <summary>
        /// Path where hublist is stored
        /// </summary>
        private const string HublistPath = "./Hublists/http___dchublist.com_hublist.xml";

        /// <summary>
        /// Address indicating where to download hublist from
        /// </summary>
        private const string HublistAddress = "http://dchublist.com/hublist.xml.bz2";

        /// <summary>
        /// Stores all file types as enum values
        /// </summary>
        private static readonly ReadOnlyCollection<SearchFileType> AllFileTypes = new ReadOnlyCollection<SearchFileType>(Enum.GetValues(typeof(SearchFileType)).OfType<SearchFileType>().ToList());
        
        /// <summary>
        /// Stores all types of size constraints as enum values
        /// </summary>
        private static readonly ReadOnlyCollection<SizeConstraintType> AllSizeConstraintTypes = new ReadOnlyCollection<SizeConstraintType>(Enum.GetValues(typeof(SizeConstraintType)).OfType<SizeConstraintType>().ToList());
        
        /// <summary>
        /// Stores all file size units as enum values
        /// </summary>
        private static readonly ReadOnlyCollection<SizeUnit> AllSizeUnits = new ReadOnlyCollection<SizeUnit>(Enum.GetValues(typeof(SizeUnit)).OfType<SizeUnit>().ToList());

        /// <summary>
        /// List of search results
        /// </summary>
        private SortableBindingList<SearchResult> allResults = new SortableBindingList<SearchResult>();

        /// <summary>
        /// Instance of DC client
        /// </summary>
        private DcClient directConnectClient;

        /// <summary>
        /// Initializes a new instance of the <see cref="MainForm" /> class
        /// </summary>
        public MainForm()
        {
            this.InitializeComponent();

            this.cmbType.DataSource = AllFileTypes;
            this.cmbSizeConstraint.DataSource = AllSizeConstraintTypes;
            this.cmbSizeUnit.DataSource = AllSizeUnits;

            this.dgvResults.AutoGenerateColumns = false;
            this.dgvResults.DataSource = this.allResults;

            this.RefreshHublistDate();

            this.txtSearchText.Select();

            this.Log("Ready.");
        }

        /// <summary>
        /// Delegate for writing to log text field
        /// </summary>
        /// <param name="text">Text to log</param>
        private delegate void LogDelegate(string text);

        /// <summary>
        /// Delegate for writing search result to result pane
        /// </summary>
        /// <param name="searchResult">Search result to write</param>
        private delegate void SearchResultDelegate(SearchResult searchResult);

        /// <summary>
        /// Delegate with no arguments
        /// </summary>
        private delegate void EmptyDelegate();

        /// <summary>
        /// Refreshes hublist date in form
        /// </summary>
        private void RefreshHublistDate()
        {
            if (File.Exists(HublistPath))
            {
                var fileTime = File.GetCreationTime(HublistPath);
                this.lblHublistDate.Text = fileTime.ToString(CultureInfo.CurrentCulture);
                this.btnUpdateHublist.Enabled = fileTime < DateTime.Now - TimeSpan.FromHours(1);
            }
            else
            {
                this.lblHublistDate.Text = "?";
            }
        }

        /// <summary>
        /// Writes a message to log text field
        /// </summary>
        /// <param name="input"></param>
        private void Log(string input)
        {
            var time = "[" + string.Format("{0:HH}", DateTime.Now) + ":" + string.Format("{0:mm}", DateTime.Now) + ":" + string.Format("{0:ss}", DateTime.Now) + "]";
            var startIndex = this.txtStatus.TextLength;
            this.txtStatus.AppendText(time + "  " + input.TrimEnd(ErrorIndicator) + Environment.NewLine);
            var endIndex = this.txtStatus.TextLength;

            if (input.Contains(ErrorIndicator))
            {
                this.txtStatus.Select(startIndex, endIndex - startIndex);
                this.txtStatus.SelectionColor = Color.Red;
            }

            this.txtStatus.ScrollToBottom();
            this.txtStatus.Update();
        }

        /// <summary>
        /// Writes a message to log text field from a different thread
        /// </summary>
        /// <param name="text"></param>
        private void InvokeLog(string text)
        {
            this.txtStatus.Invoke(new LogDelegate(this.Log), text);
        }

        /// <summary>
        /// Performs tasks when DC search was stopped
        /// </summary>
        private void OnSearchStopped()
        {
            this.btnSearch.Enabled = true;
            this.grpActive.Enabled = true;
            this.grpParameters.Enabled = true;
            this.btnStop.Enabled = false;
            this.btnSearch.Focus();
        }

        /// <summary>
        /// Starts DC search
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        [SuppressMessage("StyleCop.CSharp.NamingRules", "SA1300:ElementMustBeginWithUpperCaseLetter", Justification = "Generated handler")]
        private void btnSearch_Click(object sender, EventArgs e)
        {
            if (!StringUtility.IsIPAddress(this.txtIP.Text))
            {
                MessageBox.Show("IP address is in incorrect format (use decimal IPv4 format like X.Y.Z.W).", "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                this.txtIP.Focus();
                return;
            }

            if (string.IsNullOrWhiteSpace(this.txtSearchText.Text) || this.txtSearchText.Text.Length < 3)
            {
                MessageBox.Show("Search text cannot be empty and must be at least 3 characters long.", "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                this.txtSearchText.Focus();
                return;
            }

            if (!File.Exists(HublistPath))
            {
                MessageBox.Show("The predefined hublist does not exist.", "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                return;
            }

            this.btnSearch.Enabled = false;
            this.grpActive.Enabled = false;
            this.grpParameters.Enabled = false;
            this.btnStop.Enabled = true;
            this.btnStop.Focus();
            this.allResults.Clear();

            var searchParameters = new SearchParameters()
            {
                FileType = (SearchFileType)this.cmbType.SelectedItem,
                FreeSlotsOnly = this.chkFreeSlots.Checked,
                Size = Common.ToBytes((long)this.numSize.Value, (SizeUnit)this.cmbSizeUnit.SelectedItem),
                SizeConstraintType = (SizeConstraintType)this.cmbSizeConstraint.SelectedItem,
                Text = this.txtSearchText.Text
            };

            this.directConnectClient = new DcClient(this.txtIP.Text, (ushort)this.numPort.Value);

            this.Log("Batch DC search started.");

            this.directConnectClient.StartBatchSearch(
                File.ReadAllText("./Hublists/http___dchublist.com_hublist.xml", Encoding.UTF8),
                searchParameters,
                searchResults =>
                {
                    searchResults.ToList().ForEach(r => this.txtStatus.Invoke(new SearchResultDelegate(this.allResults.Add), r));
                },
                (hub, current, total) =>
                {
                    this.InvokeLog(
                        string.Format(
                            CultureInfo.CurrentCulture,
                            "({0}/{1}) {2}",
                            current,
                            total,
                            hub.ToString()));
                },
                errorReason =>
                {
                    this.InvokeLog(errorReason + ErrorIndicator);
                },
                () =>
                {
                    this.InvokeLog("Batch DC search finished.");
                    this.txtStatus.Invoke(new EmptyDelegate(this.OnSearchStopped));
                });
        }

        /// <summary>
        /// Stops DC search
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        [SuppressMessage("StyleCop.CSharp.NamingRules", "SA1300:ElementMustBeginWithUpperCaseLetter", Justification = "Generated handler")]
        private void btnStop_Click(object sender, EventArgs e)
        {
            if (this.directConnectClient != null)
            {
                this.directConnectClient.StopBatchSearch();
                this.InvokeLog("Batch DC search stopped.");

                this.OnSearchStopped();
            }
        }

        /// <summary>
        /// Detects IP address and shows it in appropriate field in form
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        [SuppressMessage("StyleCop.CSharp.NamingRules", "SA1300:ElementMustBeginWithUpperCaseLetter", Justification = "Generated handler")]
        private void btnDetectIP_Click(object sender, EventArgs e)
        {
            this.btnDetectIP.Enabled = false;
            IPAddressGetter.StartGetIPAddress(result =>
            {
                this.txtIP.Text = result;
                this.btnDetectIP.Enabled = true;
            });
        }

        /// <summary>
        /// Downloads fresh hublist
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        [SuppressMessage("StyleCop.CSharp.NamingRules", "SA1300:ElementMustBeginWithUpperCaseLetter", Justification = "Generated handler")]
        private void btnUpdateHublist_Click(object sender, EventArgs e)
        {
            if (MessageBox.Show(
                "This will download a new hublist from " + HublistAddress + ". Is it OK to proceed?",
                "Question",
                MessageBoxButtons.OKCancel,
                MessageBoxIcon.Question,
                MessageBoxDefaultButton.Button1) == DialogResult.OK)
            {
                HublistDownloader.StartHublistDownload(
                    HublistAddress,
                    HublistPath,
                    () =>
                    {
                        this.RefreshHublistDate();
                    });

                this.btnUpdateHublist.Enabled = false;
            }
        }

        /// <summary>
        /// Copies hub address to clipboard on doubleclicking of search result
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        [SuppressMessage("StyleCop.CSharp.NamingRules", "SA1300:ElementMustBeginWithUpperCaseLetter", Justification = "Generated handler")]
        private void dgvResults_CellContentDoubleClick(object sender, DataGridViewCellEventArgs e)
        {
            if (e.RowIndex != -1)
            {
                var address = ((SearchResult)this.dgvResults.Rows[e.RowIndex].DataBoundItem).HubAddress;
                Clipboard.Clear();
                Clipboard.SetText(address);
            }
        }
    }
}
